about = {
    "__title__": "wiji",
    "__description__": "Wiji is an asyncio distributed task processor/queue.",
    "__url__": "https://github.com/komuw/wiji",
    "__version__": "v0.1.5",
    "__author__": "komuW",
    "__author_email__": "komuw05@gmail.com",
    "__license__": "MIT",
}
